﻿using System;
using System.Collections.Generic;
using System.Threading;
using Alarus;
using Alarus.Frames.Sound;
using Alarus.Frames.Video;
using Alarus.IpDevices;
using Alarus.IpDevices.Events;
using Alarus.IpDevices.IO;
using Alarus.IpDevices.PTZ;
using Alarus.IpDevices.Storage.Archive;
using Alarus.RealTimeFrameProviding;
using Camera.Basic.Host;
using Camera.Basic.Log;

namespace Camera.Video
{
	public class StreamReceiver : ICameraServiceProvider
	{
		private readonly ConnectionParameters _conParams;

		/// <summary>
		/// Provides bytes that are already pre-allocated in memory
		/// </summary>
		private readonly IBytesMemoryManager _bytesMemoryManager;
		private readonly object _lock = new object();

		private Thread _mainVideoThread;

		public StreamReceiver(ConnectionParameters conParams, SubStreamParameters[] subParams)
		{
			_conParams = conParams;

			_bytesMemoryManager = HostFactory.Instance.CreateBytesMemoryManager();
		}

		public event NewRawFrameEventHandler NewRawFrame;
		public event NewRawEventHandler NewEvent;
		public event EventHandler NewLogRecord;

		public bool IsStreamActive(ChannelStreamTypes streamType)
		{
			return _mainVideoThread != null;
		}

		public void StartStream(ChannelStreamTypes streamType)
		{
			// Host can start several streams at the one time
			lock (_lock)
			{
				if (IsStreamActive(streamType))
					return;
				try
				{
					if (streamType != ChannelStreamTypes.MainVideo)
						return;

					_mainVideoThread = new Thread(WorkFunc)
					{
						IsBackground = true
					};
					_mainVideoThread.Start();
				}
				catch (Exception ex)
				{
					SampleLogMgr.LogException(ex, string.Empty);
				}
			}
		}

		private void WorkFunc()
		{
			try
			{
				const int fps = 25;
				const int frameDelay = 1000 / fps;

				var frameStartTime = Environment.TickCount;

				AddLogRecord(StreamTypes.Main, ChannelStreamTypes.MainVideo,
					$"Attempting to connect to the camera {ModuleInfo.DevName}({GetType().Name}) by the address: {_conParams.Address}.");

				Thread.Sleep(1000); // Simulate connection to the device logic

				AddLogRecord(StreamTypes.Main, ChannelStreamTypes.MainVideo, "Connection success. Start frames receiving.");

				InformHostAboutSuccessConnection();

				while (true)
				{
					//-----------------------------
					// Simulate gap with the camera
					//-----------------------------
					// var eventStartTime = Environment.TickCount;
					// var randomizer = new Random();
					// var disconnectDelay = 10000 + randomizer.Next(5000);
					// if (Environment.TickCount - eventStartTime >= disconnectDelay)
					// {
					//	 OnNoConnection();

					//	 Thread.Sleep(5000 + randomizer.Next(5000));
					//	 disconnectDelay = 5000 + randomizer.Next(5000);
					//	 eventStartTime = Environment.TickCount;
					//	 continue;
					// }
					//-----------------------------

					var rawMjpegFrameBytes = _bytesMemoryManager.CreateReferencedBytesFrom(Camera.Properties.Resources.testFrame);

					// Simulate the camera at a rate of 25 fps
					if (Environment.TickCount - frameStartTime < frameDelay)
						continue;

					var jpegFrame = new RawMJPEGFrame(rawMjpegFrameBytes)
					{
						TimeStamp = DateTime.UtcNow
					};

					NewRawFrame?.Invoke(this, new NewRawFrameEventArgs(ChannelStreamTypes.MainVideo, jpegFrame));
					frameStartTime = Environment.TickCount;
				}
			}
			catch (ThreadAbortException)
			{
				AddLogRecord(StreamTypes.Main, ChannelStreamTypes.MainVideo, "Manual stop of connection to the camera.");
			}
			catch (Exception ex)
			{
				AddLogRecord(StreamTypes.Main, ChannelStreamTypes.MainVideo, "* Error while frames receiving: " + ex.Message);
				SampleLogMgr.LogException(ex, string.Empty);
			}
		}

		private void InformHostAboutSuccessConnection()
		{
			var establishedPhysicalConnectionDeviceEvent = new EstablishedPhysicalConnectionDeviceEvent
			{
				StreamTypesMask = ChannelStreamTypes.MainVideo
			};
			NewEvent?.Invoke(this, new NewEventEventArgs(establishedPhysicalConnectionDeviceEvent));
		}

		public void StopStream(ChannelStreamTypes streamType)
		{
			try
			{
				if (streamType == ChannelStreamTypes.MainVideo)
				{
					if (_mainVideoThread != null)
					{
						_mainVideoThread.Abort();
						_mainVideoThread = null;
					}
				}
			}
			catch (Exception ex)
			{
				SampleLogMgr.LogException(ex, string.Empty);
			}
		}

		public void Release()
		{
			StopStream(ChannelStreamTypes.MainVideo);
		}

		public void SendSound(RawSoundFrame frame)
		{
		}

		private void OnNoConnection()
		{
			var disEvent = new NoDataConnectionDeviceEvent
			{
				StreamTypesMask = ChannelStreamTypes.MainVideo
			};

			NewEvent?.Invoke(this, new NewEventEventArgs(disEvent));
		}

		public IAsyncResult BeginGetCapabilities(AsyncCallback callback, object state)
		{
			throw new NotImplementedException();
		}

		public DeviceCapabilities EndGetCapabilities(IAsyncResult asyncResult)
		{
			throw new NotImplementedException();
		}

		public IPtzController GetPtzController()
		{
			throw new NotImplementedException();
		}

		public IDigitalOutputsController GetDigitalOutputsController()
		{
			throw new NotImplementedException();
		}

		public IDeviceArchiveController GetDeviceArchiveController()
		{
			throw new NotImplementedException();
		}

		#region Camera connection log

		private int maxLinesInConnectionLog = 100;

		private bool isWritingConnectionLog;

		public bool IsWritingConnectionLog
		{
			get { return isWritingConnectionLog; }
			set { isWritingConnectionLog = value; }
		}

		private List<string> connectionLog = new List<string>();

		public string ConnectionLog
		{
			get
			{
				lock (connectionLog)
				{
					string logLine = string.Empty;

					foreach (string line in connectionLog)
						logLine += line + "\r\n";

					return logLine;
				}
			}
		}

		public void ClearConnectionLog()
		{
			lock (connectionLog)
			{
				connectionLog.Clear();
			}
		}

		public void AddLogRecord(StreamTypes streamType, ChannelStreamTypes channelStreamType, string info)
		{
			if (!isWritingConnectionLog)
				return;

			string typeOfData = string.Empty;

			switch (channelStreamType)
			{
				case ChannelStreamTypes.MainVideo:
				case ChannelStreamTypes.AlternativeVideo:
					typeOfData = "video";
					break;
				case ChannelStreamTypes.MainSound:
				case ChannelStreamTypes.AlternativeSound:
					typeOfData = "sound";
					break;
				case ChannelStreamTypes.MotionDetection:
					typeOfData = "motion detection";
					break;
			}

			string streamTypeName = string.Empty;

			if (streamType == StreamTypes.Main)
				streamTypeName = "main stream";
			else
				streamTypeName = "alternative stream";

			string str = string.Format("[{0}, {1}, {2}] {3}", DateTime.Now.ToLongTimeString(), typeOfData, streamTypeName, info);

			lock (connectionLog)
			{
				connectionLog.Add(str);

				if (connectionLog.Count > maxLinesInConnectionLog)
					connectionLog.RemoveAt(0);
			}

			NewLogRecord(this, new EventArgs());
		}

		public void AddLogRecord(StreamTypes subType, ChannelStreamTypes subChannelType, string format, params object[] args)
		{
			AddLogRecord(subType, subChannelType, string.Format(format, args));
		}

		#endregion
	}
}
